@@ -15,6 +15,7 @@ import shutil
 import subprocess
 import tempfile
 import urllib.parse
+import warnings
 
 from gemato.exceptions import (
     OpenPGPNoImplementation,
@@ -22,8 +23,10 @@ from gemato.exceptions import (
     OpenPGPExpiredKeyFailure,
     OpenPGPRevokedKeyFailure,
     OpenPGPKeyImportError,
+    OpenPGPKeyListingError,
     OpenPGPKeyRefreshError,
     OpenPGPUnknownSigFailure,
+    OpenPGPUntrustedSigFailure,
     OpenPGPSigningFailure,
     )
 
@@ -32,6 +35,11 @@ try:
 except ImportError:
     requests = None
 
+try:
+    import pgpy
+except ImportError:
+    pgpy = None
+
 
 GNUPG = os.environ.get('GNUPG', 'gpg')
 GNUPGCONF = os.environ.get('GNUPGCONF', 'gpgconf')
@@ -49,7 +57,22 @@ class OpenPGPSignatureData:
         self.primary_key_fingerprint = primary_key_fingerprint
 
 
-class OpenPGPSystemEnvironment:
+ZBASE32_TRANSLATE = bytes.maketrans(
+    b'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567',
+    b'ybndrfg8ejkmcpqxot1uwisza345h769')
+
+
+def get_wkd_url(email):
+    localname, domain = email.split('@', 1)
+    b32 = (
+        base64.b32encode(
+            hashlib.sha1(localname.encode('utf8').lower()).digest())
+        .translate(ZBASE32_TRANSLATE).decode('ASCII'))
+    return (f'https://{domain.lower()}/.well-known/openpgpkey/hu/'
+            f'{b32}?l={urllib.parse.quote(localname)}')
+
+
+class SystemGPGEnvironment:
     """
     OpenPGP environment class that uses the global OpenPGP environment
     (user's home directory or GNUPGHOME).
@@ -69,7 +92,7 @@ class OpenPGPSystemEnvironment:
     def close(self):
         pass
 
-    def import_key(self, keyfile):
+    def import_key(self, keyfile, trust=True):
         """
         Import a public key from open file @keyfile. The file should
         be open for reading in binary mode, and oriented
@@ -121,11 +144,11 @@ class OpenPGPSystemEnvironment:
 
         exitst, out, err = self._spawn_gpg(
             [GNUPG, '--batch', '--status-fd', '1', '--verify'],
-            f.read().encode('utf8'))
-        if exitst != 0:
-            raise OpenPGPVerificationFailure(err.decode('utf8'))
+            f.read().encode('utf8'),
+            raise_on_error=OpenPGPVerificationFailure)
 
         is_good = False
+        is_trusted = False
         sig_data = None
 
         # process the output of gpg to find the exact result
@@ -133,9 +156,11 @@ class OpenPGPSystemEnvironment:
             if line.startswith(b'[GNUPG:] GOODSIG'):
                 is_good = True
             elif line.startswith(b'[GNUPG:] EXPKEYSIG'):
-                raise OpenPGPExpiredKeyFailure(err.decode('utf8'))
+                raise OpenPGPExpiredKeyFailure(
+                    err.decode('utf8', errors='backslashreplace'))
             elif line.startswith(b'[GNUPG:] REVKEYSIG'):
-                raise OpenPGPRevokedKeyFailure(err.decode('utf8'))
+                raise OpenPGPRevokedKeyFailure(
+                    err.decode('utf8', errors='backslashreplace'))
             elif line.startswith(b'[GNUPG:] VALIDSIG'):
                 spl = line.split(b' ')
                 assert len(spl) >= 12
@@ -145,10 +170,20 @@ class OpenPGPSystemEnvironment:
                 pkfp = spl[11].decode('utf8')
 
                 sig_data = OpenPGPSignatureData(fp, ts, expts, pkfp)
+            elif line.startswith(b'[GNUPG:] TRUST_'):
+                spl = line.split(b' ', 2)
+                if spl[1] in (b'TRUST_MARGINAL',
+                              b'TRUST_FULL',
+                              b'TRUST_ULTIMATE'):
+                    is_trusted = True
 
         # require both GOODSIG and VALIDSIG
         if not is_good or sig_data is None:
-            raise OpenPGPUnknownSigFailure(err.decode('utf8'))
+            raise OpenPGPUnknownSigFailure(
+                err.decode('utf8', errors='backslashreplace'))
+        if not is_trusted:
+            raise OpenPGPUntrustedSigFailure(
+                err.decode('utf8', errors='backslashreplace'))
         return sig_data
 
     def clear_sign_file(self, f, outf, keyid=None):
@@ -167,13 +202,13 @@ class OpenPGPSystemEnvironment:
             args += ['--local-user', keyid]
         exitst, out, err = self._spawn_gpg(
             [GNUPG, '--batch', '--clearsign'] + args,
-            f.read().encode('utf8'))
-        if exitst != 0:
-            raise OpenPGPSigningFailure(err.decode('utf8'))
+            f.read().encode('utf8'),
+            raise_on_error=OpenPGPSigningFailure)
 
         outf.write(out.decode('utf8'))
 
-    def _spawn_gpg(self, argv, stdin='', env_override={}):
+    def _spawn_gpg(self, argv, stdin='', env_override={},
+                   raise_on_error=None):
         env = os.environ.copy()
         env['TZ'] = 'UTC'
         env.update(env_override)
@@ -184,16 +219,17 @@ class OpenPGPSystemEnvironment:
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  env=env)
-        except OSError as e:
-            if e.errno != errno.ENOENT:
-                raise
-            raise OpenPGPNoImplementation()
+        except FileNotFoundError:
+            raise OpenPGPNoImplementation('install gpg')
 
         out, err = p.communicate(stdin)
+        if raise_on_error is not None and p.wait() != 0:
+            raise raise_on_error(
+                err.decode('utf8', errors='backslashreplace'))
         return (p.wait(), out, err)
 
 
-class OpenPGPEnvironment(OpenPGPSystemEnvironment):
+class IsolatedGPGEnvironment(SystemGPGEnvironment):
     """
     An isolated environment for OpenPGP routines. Used to get reliable
     verification results independently of user configuration.
@@ -222,8 +258,8 @@ debug-level guru
         with open(os.path.join(self._home, 'gpg.conf'), 'w') as f:
             f.write('''# autogenerated by gemato
 
-# we are using an isolated keyring, so always trust our keys
-trust-model always
+# we set validity directly on keys
+trust-model direct
 ''')
         with open(os.path.join(self._home, 'gpg-agent.conf'), 'w') as f:
             f.write(f'''# autogenerated by gemato
@@ -241,7 +277,7 @@ debug-level guru
             self.close()
 
     def clone(self):
-        return OpenPGPEnvironment(debug=self.debug, proxy=self.proxy)
+        return IsolatedGPGEnvironment(debug=self.debug, proxy=self.proxy)
 
     @staticmethod
     def _rmtree_error_handler(func, path, exc_info):
@@ -261,7 +297,9 @@ debug-level guru
             ret, sout, serr = self._spawn_gpg(
                 [GNUPGCONF, '--kill', 'all'])
             if ret != 0:
-                logging.warning(f'{GNUPGCONF} --kill failed: {serr}')
+                logging.warning(
+                    f'{GNUPGCONF} --kill failed:\n'
+                    f'{serr.decode("utf8", errors="backslashescape")}')
             if not self.debug:
                 # we need to loop due to ENOTEMPTY potential
                 while os.path.isdir(self._home):
@@ -272,26 +310,77 @@ debug-level guru
                               f'{self._home}')
             self._home = None
 
-    def import_key(self, keyfile):
+    def import_key(self, keyfile, trust=True):
         exitst, out, err = self._spawn_gpg(
-            [GNUPG, '--batch', '--import'], keyfile.read())
-        if exitst != 0:
-            raise OpenPGPKeyImportError(err.decode('utf8'))
+            [GNUPG, '--batch', '--import', '--status-fd', '1'],
+            keyfile.read(),
+            raise_on_error=OpenPGPKeyImportError)
 
-    zbase32_translate = bytes.maketrans(
-        b'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567',
-        b'ybndrfg8ejkmcpqxot1uwisza345h769')
+        if trust:
+            fprs = set()
+            for line in out.splitlines():
+                if line.startswith(b'[GNUPG:] IMPORT_OK'):
+                    fprs.add(line.split(b' ')[3].decode('ASCII'))
 
-    @classmethod
-    def get_wkd_url(cls, email):
-        localname, domain = email.encode('utf8').split(b'@', 1)
-        b32 = (
-            base64.b32encode(hashlib.sha1(localname.lower()).digest())
-            .translate(cls.zbase32_translate).decode())
-        uenc = urllib.parse.quote(localname)
-        ldomain = domain.lower().decode('utf8')
-        return (f'https://{ldomain}/.well-known/openpgpkey/hu/'
-                f'{b32}?l={uenc}')
+            ownertrust = ''.join(f'{fpr}:6:\n' for fpr in fprs).encode('utf8')
+            exitst, out, err = self._spawn_gpg(
+                [GNUPG, '--batch', '--import-ownertrust'],
+                ownertrust,
+                raise_on_error=OpenPGPKeyImportError)
+
+    def list_keys(self):
+        """
+        List fingerprints and UIDs of all keys in keyring
+
+        Returns a mapping from fingerprint (as a string) to an iterable
+        of UIDs.
+        """
+
+        exitst, out, err = self._spawn_gpg(
+            [GNUPG, '--batch', '--with-colons', '--list-keys'],
+            raise_on_error=OpenPGPKeyListingError)
+
+        prev_pub = None
+        fpr = None
+        ret = {}
+
+        for line in out.splitlines():
+            # were we expecting a fingerprint?
+            if prev_pub is not None:
+                if line.startswith(b'fpr:'):
+                    fpr = line.split(b':')[9].decode('ASCII')
+                    if not fpr.endswith(prev_pub):
+                        raise OpenPGPKeyListingError(
+                            f'Incorrect fingerprint {fpr} for key '
+                            f'{prev_pub}')
+                    logging.debug(
+                        f'list_keys(): fingerprint: {fpr}')
+                    ret[fpr] = []
+                    prev_pub = None
+                else:
+                    raise OpenPGPKeyListingError(
+                        f'No fingerprint in GPG output, instead got: '
+                        f'{line}')
+            elif line.startswith(b'pub:'):
+                # wait for the fingerprint
+                prev_pub = line.split(b':')[4].decode('ASCII')
+                logging.debug(f'list_keys(): keyid: {prev_pub}')
+            elif line.startswith(b'uid:'):
+                if fpr is None:
+                    raise OpenPGPKeyListingError(
+                        f'UID without key in GPG output: {line}')
+                uid = line.split(b':')[9]
+                _, addr = email.utils.parseaddr(
+                    uid.decode('utf8', errors='replace'))
+                if '@' in addr:
+                    logging.debug(f'list_keys(): UID: {addr}')
+                    ret[fpr].append(addr)
+                else:
+                    logging.debug(
+                        f'list_keys(): ignoring UID without mail: '
+                        f'{uid!r}')
+
+        return ret
 
     def refresh_keys_wkd(self):
         """
@@ -304,61 +393,19 @@ debug-level guru
             return False
 
         # list all keys in the keyring
-        exitst, out, err = self._spawn_gpg(
-            [GNUPG, '--batch', '--with-colons', '--list-keys'])
-        if exitst != 0:
-            raise OpenPGPKeyRefreshError(err.decode('utf8'))
-
-        # find keys and UIDs
-        addrs = set()
-        addrs_key = set()
-        keys = set()
-        prev_pub = None
-        for line in out.splitlines():
-            # were we expecting a fingerprint?
-            if prev_pub is not None:
-                if line.startswith(b'fpr:'):
-                    fpr = line.split(b':')[9].decode('ASCII')
-                    assert fpr.endswith(prev_pub)
-                    logging.debug(
-                        f'refresh_keys_wkd(): fingerprint: {fpr}')
-                    keys.add(fpr)
-                    prev_pub = None
-                else:
-                    # old GnuPG doesn't give fingerprints by default
-                    # (but it doesn't support WKD either)
-                    logging.debug(
-                        'refresh_keys_wkd(): failing due to old gpg')
-                    return False
-            elif line.startswith(b'pub:'):
-                if keys:
-                    # every key must have at least one UID
-                    if not addrs_key:
-                        logging.debug(
-                            'refresh_keys_wkd(): failing due to no UIDs')
-                        return False
-                    addrs.update(addrs_key)
-                    addrs_key = set()
-
-                # wait for the fingerprint
-                prev_pub = line.split(b':')[4].decode('ASCII')
-                logging.debug(f'refresh_keys_wkd(): keyid: {prev_pub}')
-            elif line.startswith(b'uid:'):
-                uid = line.split(b':')[9]
-                name, addr = email.utils.parseaddr(uid.decode('utf8'))
-                if '@' in addr:
-                    logging.debug(f'refresh_keys_wkd(): UID: {addr}')
-                    addrs_key.add(addr)
-                else:
-                    logging.debug(
-                        f'refresh_keys_wkd(): ignoring UID without '
-                        f'mail: {uid.decode("utf8")}')
-
-        # grab the final set (also aborts when there are no keys)
-        if not addrs_key:
-            logging.debug('refresh_keys_wkd(): failing due to no UIDs')
+        keys = self.list_keys()
+        if not keys:
+            logging.debug('refresh_keys_wkd(): no keys found')
             return False
-        addrs.update(addrs_key)
+        addrs = set()
+        for key, uids in keys.items():
+            if not uids:
+                logging.debug(
+                    f'refresh_keys_wkd(): failing due to no UIDs on '
+                    f'key {key}')
+                return False
+            addrs.update(uids)
+        keys = set(keys)
 
         data = b''
         proxies = {}
@@ -368,7 +415,7 @@ debug-level guru
                 'https': self.proxy,
             }
         for a in addrs:
-            url = self.get_wkd_url(a)
+            url = get_wkd_url(a)
             resp = requests.get(url, proxies=proxies)
             if resp.status_code != 200:
                 logging.debug(f'refresh_keys_wkd(): failing due to failed'
@@ -377,10 +424,9 @@ debug-level guru
             data += resp.content
 
         exitst, out, err = self._spawn_gpg(
-            [GNUPG, '--batch', '--import', '--status-fd', '1'], data)
-        if exitst != 0:
-            # there's no valid reason for import to fail here
-            raise OpenPGPKeyRefreshError(err.decode('utf8'))
+            [GNUPG, '--batch', '--import', '--status-fd', '1'],
+            data,
+            raise_on_error=OpenPGPKeyRefreshError)
 
         # we need to explicitly ensure all keys were fetched
         for line in out.splitlines():
@@ -393,10 +439,8 @@ debug-level guru
                 else:
                     # we need to delete unexpected keys
                     exitst, out, err = self._spawn_gpg(
-                        [GNUPG, '--batch', '--delete-keys', fpr])
-                    if exitst != 0:
-                        raise OpenPGPKeyRefreshError(
-                            err.decode('utf8'))
+                        [GNUPG, '--batch', '--delete-keys', fpr],
+                        raise_on_error=OpenPGPKeyRefreshError)
         if keys:
             logging.debug(
                 f'refresh_keys_wkd(): failing due to non-updated keys: '
@@ -411,9 +455,8 @@ debug-level guru
             ks_args = ['--keyserver', keyserver]
 
         exitst, out, err = self._spawn_gpg(
-            [GNUPG, '--batch', '--refresh-keys'] + ks_args)
-        if exitst != 0:
-            raise OpenPGPKeyRefreshError(err.decode('utf8'))
+            [GNUPG, '--batch', '--refresh-keys'] + ks_args,
+            raise_on_error=OpenPGPKeyRefreshError)
 
     def refresh_keys(self, allow_wkd=True, keyserver=None):
         logging.debug(f'refresh_keys(allow_wkd={allow_wkd}, '
@@ -429,8 +472,119 @@ debug-level guru
         assert self._home is not None
         return self._home
 
-    def _spawn_gpg(self, options, stdin=''):
+    def _spawn_gpg(self, *args, **kwargs):
         env_override = {'GNUPGHOME': self.home}
         if self.proxy is not None:
             env_override['http_proxy'] = self.proxy
-        return (super()._spawn_gpg(options, stdin, env_override))
+        assert 'env_override' not in kwargs
+        kwargs['env_override'] = env_override
+        return super()._spawn_gpg(*args, **kwargs)
+
+
+class PGPyEnvironment:
+    """Stand-alone environment using pgpy library"""
+
+    __slots__ = ['debug', 'keyring', 'proxy']
+
+    def __init__(self, debug=False, proxy=None):
+        if pgpy is None:
+            raise OpenPGPNoImplementation('install PGPy')
+        self.debug = debug
+        self.keyring = pgpy.PGPKeyring()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, exc_cb):
+        pass
+
+    def close(self):
+        pass
+
+    def import_key(self, keyfile):
+        with warnings.catch_warnings(record=True) as warns:
+            try:
+                key_res = pgpy.PGPKey.from_blob(keyfile.read())
+            except ValueError as e:
+                raise OpenPGPKeyImportError(
+                    f'OpenPGP key import failed: {e}')
+        fprs = []
+        for k in key_res[1].values():
+            fprs.extend(self.keyring.load(k))
+
+        for w in warns:
+            if str(w.message) == 'Incorrect crc24':
+                raise OpenPGPKeyImportError(
+                    f'OpenPGP key import failed: {w.message}')
+
+        for fpr in fprs:
+            with self.keyring.key(fpr) as k:
+                if k.parent is not None:
+                    try:
+                        verifies = k.parent.verify(k)
+                    except pgpy.errors.PGPError:
+                        logging.debug(
+                            f'Rejecting subkey {fpr} due to missing sig')
+                        self.keyring.unload(k)
+                    else:
+                        if not verifies:
+                            logging.debug(
+                                f'Rejecting subkey {fpr} since parent '
+                                f'key signature does not check out')
+                            self.keyring.unload(k)
+                for uid in k.userids:
+                    if uid.selfsig is None:
+                        raise OpenPGPKeyImportError(
+                            f'Self-signature on {uid} missing')
+                    if not k.verify(uid):
+                        raise OpenPGPKeyImportError(
+                            f'Self-signature on {uid} does not verify')
+
+    def verify_file(self, f):
+        msg = pgpy.PGPMessage.from_blob(f.read())
+        assert msg.is_signed
+        assert len(msg.signatures) == 1
+        assert len(msg.signers) == 1
+
+        signer, = msg.signers
+        try:
+            with self.keyring.key(signer) as k:
+                pk = k
+                if k.parent is not None:
+                    pk = k.parent
+                assert pk.parent is None
+
+                vr = k.verify(msg)
+                if not vr:
+                    raise OpenPGPVerificationFailure(
+                        f'Bad signature made by key {k.fingerprint}')
+                now = datetime.datetime.utcnow()
+                sig_expire = msg.signatures[0].expires_at
+                if sig_expire is not None and sig_expire < now:
+                    raise OpenPGPVerificationFailure(
+                        f'Signature expired at {msg.signatures[0].expires_at}')
+                if k.expires_at is not None and k.expires_at < now:
+                    raise OpenPGPExpiredKeyFailure(
+                        f'Key {k.fingerprint} expired at {k.expires_at}')
+                if pk.expires_at is not None and pk.expires_at < now:
+                    raise OpenPGPExpiredKeyFailure(
+                        f'Primary key {pk.fingerprint} expired '
+                        f'at {k.expires_at}')
+                if list(k.revocation_signatures):
+                    raise OpenPGPRevokedKeyFailure(
+                        f'Key {pk.fingerprint} was revoked')
+                if list(pk.revocation_signatures):
+                    raise OpenPGPRevokedKeyFailure(
+                        f'Primary key {pk.fingerprint} was revoked')
+                return OpenPGPSignatureData(
+                    k.fingerprint,
+                    msg.signatures[0].created,
+                    msg.signatures[0].expires_at,
+                    pk.fingerprint)
+        except KeyError:
+            raise OpenPGPVerificationFailure(
+                f'Key {signer} not in keyring')
+
+
+OpenPGPSystemEnvironment = SystemGPGEnvironment
+OpenPGPEnvironment = IsolatedGPGEnvironment
