@@ -3,13 +3,13 @@
 # (c) 2017-2020 Michał Górny
 # Licensed under the terms of 2-clause BSD license
 
-import errno
 import os.path
 
 from gemato.compression import (
     open_potentially_compressed_path,
     get_potential_compressed_names,
     get_compressed_suffix_from_filename,
+    InvalidCompressedFileExceptions,
     )
 from gemato.exceptions import (
     ManifestMismatch,
@@ -322,22 +322,21 @@ class ManifestRecursiveLoader:
         try:
             m, st = self.manifest_loader.verify_and_load(
                     relpath, verify_entry)
-        except IOError as err:
-            if err.errno == errno.ENOENT and allow_create:
-                m = ManifestFile()
-                path = os.path.join(self.root_directory, relpath)
-                st = os.stat(os.path.dirname(path))
-                # trigger saving
-                self.updated_manifests.add(relpath)
+        except FileNotFoundError:
+            if not allow_create:
+                raise
+            m = ManifestFile()
+            path = os.path.join(self.root_directory, relpath)
+            st = os.stat(os.path.dirname(path))
+            # trigger saving
+            self.updated_manifests.add(relpath)
 
-                # add initial IGNORE entries to top-level Manifest
-                if relpath == 'Manifest':
-                    for ip in (self.profile
-                               .get_ignore_paths_for_new_manifest('')):
-                        ie = ManifestEntryIGNORE(ip)
-                        m.entries.append(ie)
-            else:
-                raise err
+            # add initial IGNORE entries to top-level Manifest
+            if relpath == 'Manifest':
+                for ip in (self.profile
+                           .get_ignore_paths_for_new_manifest('')):
+                    ie = ManifestEntryIGNORE(ip)
+                    m.entries.append(ie)
 
         if store_dev:
             self.manifest_device = st.st_dev
@@ -374,6 +373,7 @@ class ManifestRecursiveLoader:
                    sort=sort,
                    openpgp_env=self.openpgp_env,
                    openpgp_keyid=self.openpgp_keyid)
+            f.flush()
             return f.buffer.tell()
 
     def _iter_unordered_manifests_for_path(self, path, recursive=False):
@@ -1069,6 +1069,14 @@ class ManifestRecursiveLoader:
                     except ManifestSyntaxError:
                         # syntax error? probably not a Manifest then.
                         pass
+                    except OSError as exc:
+                        # bz2 returns generic OSError without errno
+                        # so non-null errno probably means something
+                        # else happened
+                        if exc.errno is not None:
+                            raise
+                    except InvalidCompressedFileExceptions:
+                        pass
                     else:
                         new_manifests.append(fpath)
 
