@@ -6,6 +6,8 @@
 import datetime
 import io
 import os
+import shlex
+import signal
 import tempfile
 
 import pytest
@@ -22,17 +24,21 @@ from gemato.exceptions import (
     OpenPGPKeyImportError,
     OpenPGPKeyRefreshError,
     OpenPGPRuntimeError,
+    OpenPGPUntrustedSigFailure,
     )
 from gemato.manifest import ManifestFile
 from gemato.openpgp import (
-    GNUPG,
-    OpenPGPEnvironment,
-    OpenPGPSystemEnvironment,
+    SystemGPGEnvironment,
+    IsolatedGPGEnvironment,
+    PGPyEnvironment,
+    get_wkd_url,
     )
 from gemato.recursiveloader import ManifestRecursiveLoader
 
 from tests.keydata import (
     PUBLIC_KEY, SECRET_KEY, PUBLIC_SUBKEY, UID,
+    UID_NOEMAIL, PUBLIC_KEY_NOEMAIL_SIG,
+    UID_NONUTF, PUBLIC_KEY_NONUTF_SIG,
     PUBLIC_KEY_SIG, PUBLIC_SUBKEY_SIG, EXPIRED_KEY_SIG, REVOCATION_SIG,
     OTHER_PUBLIC_KEY, OTHER_PUBLIC_KEY_UID, OTHER_PUBLIC_KEY_SIG,
     UNEXPIRE_SIG,
@@ -183,6 +189,9 @@ OTHER_VALID_PUBLIC_KEY = (OTHER_PUBLIC_KEY + OTHER_PUBLIC_KEY_UID +
                           OTHER_PUBLIC_KEY_SIG)
 OTHER_KEY_FINGERPRINT = '4B8349B90C56EE7F054D52871822F5424EB6DA81'
 
+VALID_KEY_NOEMAIL = PUBLIC_KEY + UID_NOEMAIL + PUBLIC_KEY_NOEMAIL_SIG
+VALID_KEY_NONUTF = PUBLIC_KEY + UID_NONUTF + PUBLIC_KEY_NONUTF_SIG
+
 VALID_KEY_SUBKEY = (PUBLIC_KEY + UID + PUBLIC_KEY_SIG + PUBLIC_SUBKEY +
                     PUBLIC_SUBKEY_SIG)
 SUBKEY_FINGERPRINT = '7E9DDE3CBE47E437418DF74038B9D2F76CC833CC'
@@ -227,6 +236,8 @@ FORGED_UNEXPIRE_KEY = (PUBLIC_KEY + UID + EXPIRED_KEY_SIG +
 UNSIGNED_PUBLIC_KEY = PUBLIC_KEY + UID
 UNSIGNED_SUBKEY = PUBLIC_KEY + UID + PUBLIC_KEY_SIG + PUBLIC_SUBKEY
 
+COMBINED_PUBLIC_KEYS = OTHER_VALID_PUBLIC_KEY + VALID_PUBLIC_KEY
+
 
 def strip_openpgp(text):
     lines = text.lstrip().splitlines()
@@ -322,36 +333,45 @@ def test_noverify_load_cli(tmp_path):
                                  '--no-openpgp-verify', str(tmp_path)])
 
 
-class OpenPGPMockedSystemEnvironment(OpenPGPSystemEnvironment):
+class MockedSystemGPGEnvironment(SystemGPGEnvironment):
     """System environment variant mocked to use isolated GNUPGHOME"""
     def __init__(self, *args, **kwargs):
         self._tmpdir = tempfile.TemporaryDirectory()
         os.environ['GNUPGHOME'] = self._tmpdir.name
         super().__init__(*args, **kwargs)
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, exc_cb):
+        self.close()
+
     def close(self):
         if self._tmpdir is not None:
             self._tmpdir.cleanup()
             self._tmpdir = None
             os.environ.pop('GNUPGHOME', None)
 
-    def import_key(self, keyfile):
-        exitst, out, err = self._spawn_gpg(
-            [GNUPG, '--batch', '--import'], keyfile.read())
-        if exitst != 0:
-            raise OpenPGPKeyImportError(err.decode('utf8'))
+    def import_key(self, keyfile, trust=True):
+        IsolatedGPGEnvironment.import_key(self, keyfile, trust=trust)
 
 
-@pytest.fixture(params=[OpenPGPEnvironment,
-                        OpenPGPMockedSystemEnvironment])
+@pytest.fixture(params=[IsolatedGPGEnvironment,
+                        MockedSystemGPGEnvironment,
+                        PGPyEnvironment,
+                        ])
 def openpgp_env(request):
     """OpenPGP environment fixture"""
-    env = request.param()
+    try:
+        env = request.param()
+    except OpenPGPNoImplementation as e:
+        pytest.skip(str(e))
     yield env
     env.close()
 
 
-@pytest.fixture(params=[OpenPGPEnvironment])
+@pytest.fixture(params=[IsolatedGPGEnvironment,
+                        ])
 def openpgp_env_with_refresh(request):
     """OpenPGP environments that support refreshing keys"""
     env = request.param()
@@ -363,6 +383,9 @@ MANIFEST_VARIANTS = [
     # manifest, key, expected fpr/exception
     # == good manifests ==
     ('SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', None),
+    ('SIGNED_MANIFEST', 'VALID_KEY_NOEMAIL', None),
+    ('SIGNED_MANIFEST', 'VALID_KEY_NONUTF', None),
+    ('SIGNED_MANIFEST', 'COMBINED_PUBLIC_KEYS', None),
     ('DASH_ESCAPED_SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', None),
     ('SUBKEY_SIGNED_MANIFEST', 'VALID_KEY_SUBKEY', None),
     # == using private key ==
@@ -379,6 +402,16 @@ MANIFEST_VARIANTS = [
      OpenPGPExpiredKeyFailure),
     ('SIGNED_MANIFEST', 'REVOKED_PUBLIC_KEY',
      OpenPGPRevokedKeyFailure),
+    ('SIGNED_MANIFEST', 'OTHER_VALID_PUBLIC_KEY',
+     OpenPGPVerificationFailure),
+    ('SIGNED_MANIFEST', 'UNSIGNED_PUBLIC_KEY',
+     OpenPGPKeyImportError),
+    ('SIGNED_MANIFEST', 'FORGED_PUBLIC_KEY',
+     OpenPGPKeyImportError),
+    ('SUBKEY_SIGNED_MANIFEST', 'UNSIGNED_SUBKEY',
+     OpenPGPVerificationFailure),
+    ('SUBKEY_SIGNED_MANIFEST', 'FORGED_SUBKEY',
+     OpenPGPVerificationFailure),
 ]
 
 
@@ -400,17 +433,38 @@ def assert_signature(sig, manifest_var):
                          MANIFEST_VARIANTS)
 def test_verify_manifest(openpgp_env, manifest_var, key_var, expected):
     """Test direct Manifest data verification"""
-    try:
-        if key_var is not None:
-            with io.BytesIO(globals()[key_var]) as f:
-                openpgp_env.import_key(f)
+    if (isinstance(openpgp_env, PGPyEnvironment) and
+            manifest_var == 'DASH_ESCAPED_SIGNED_MANIFEST'):
+        pytest.xfail('dash escaping is known-broken in pgpy')
 
+    try:
         with io.StringIO(globals()[manifest_var]) as f:
             if expected is None:
+                if key_var is not None:
+                    with io.BytesIO(globals()[key_var]) as kf:
+                        openpgp_env.import_key(kf)
+
                 sig = openpgp_env.verify_file(f)
                 assert_signature(sig, manifest_var)
             else:
                 with pytest.raises(expected):
+                    if key_var is not None:
+                        with io.BytesIO(globals()[key_var]) as kf:
+                            openpgp_env.import_key(kf)
+
+                    openpgp_env.verify_file(f)
+    except OpenPGPNoImplementation as e:
+        pytest.skip(str(e))
+
+
+def test_verify_untrusted_key():
+    try:
+        with MockedSystemGPGEnvironment() as openpgp_env:
+            with io.BytesIO(VALID_PUBLIC_KEY) as f:
+                openpgp_env.import_key(f, trust=False)
+
+            with io.StringIO(SIGNED_MANIFEST) as f:
+                with pytest.raises(OpenPGPUntrustedSigFailure):
                     openpgp_env.verify_file(f)
     except OpenPGPNoImplementation as e:
         pytest.skip(str(e))
@@ -420,26 +474,38 @@ def test_verify_manifest(openpgp_env, manifest_var, key_var, expected):
                          MANIFEST_VARIANTS)
 def test_manifest_load(openpgp_env, manifest_var, key_var, expected):
     """Test Manifest verification via ManifestFile.load()"""
-    try:
-        if key_var is not None:
-            with io.BytesIO(globals()[key_var]) as f:
-                openpgp_env.import_key(f)
+    if (isinstance(openpgp_env, PGPyEnvironment) and
+            manifest_var == 'DASH_ESCAPED_SIGNED_MANIFEST'):
+        pytest.xfail('dash escaping is known-broken in pgpy')
 
+    try:
+        key_loaded = False
         m = ManifestFile()
         with io.StringIO(globals()[manifest_var]) as f:
             if expected is None:
+                if key_var is not None:
+                    with io.BytesIO(globals()[key_var]) as kf:
+                        openpgp_env.import_key(kf)
+
+                key_loaded = True
                 m.load(f, openpgp_env=openpgp_env)
                 assert m.openpgp_signed
                 assert_signature(m.openpgp_signature, manifest_var)
             else:
                 with pytest.raises(expected):
+                    if key_var is not None:
+                        with io.BytesIO(globals()[key_var]) as kf:
+                            openpgp_env.import_key(kf)
+
+                    key_loaded = True
                     m.load(f, openpgp_env=openpgp_env)
                 assert not m.openpgp_signed
                 assert m.openpgp_signature is None
 
-        # Manifest entries should be loaded even if verification failed
-        assert m.find_timestamp() is not None
-        assert m.find_path_entry('myebuild-0.ebuild') is not None
+        if key_loaded:
+            # Manifest entries should be loaded even if verification failed
+            assert m.find_timestamp() is not None
+            assert m.find_path_entry('myebuild-0.ebuild') is not None
     except OpenPGPNoImplementation as e:
         pytest.skip(str(e))
 
@@ -450,15 +516,19 @@ def test_manifest_load(openpgp_env, manifest_var, key_var, expected):
 def test_recursive_manifest_loader(tmp_path, openpgp_env, filename,
                                    manifest_var, key_var, expected):
     """Test Manifest verification via ManifestRecursiveLoader"""
-    try:
-        if key_var is not None:
-            with io.BytesIO(globals()[key_var]) as f:
-                openpgp_env.import_key(f)
+    if (isinstance(openpgp_env, PGPyEnvironment) and
+            manifest_var == 'DASH_ESCAPED_SIGNED_MANIFEST'):
+        pytest.xfail('dash escaping is known-broken in pgpy')
 
+    try:
         with open_potentially_compressed_path(tmp_path / filename, 'w') as cf:
             cf.write(globals()[manifest_var])
 
         if expected is None:
+            if key_var is not None:
+                with io.BytesIO(globals()[key_var]) as f:
+                    openpgp_env.import_key(f)
+
             m = ManifestRecursiveLoader(tmp_path / filename,
                                         verify_openpgp=True,
                                         openpgp_env=openpgp_env)
@@ -466,6 +536,10 @@ def test_recursive_manifest_loader(tmp_path, openpgp_env, filename,
             assert_signature(m.openpgp_signature, manifest_var)
         else:
             with pytest.raises(expected):
+                if key_var is not None:
+                    with io.BytesIO(globals()[key_var]) as f:
+                        openpgp_env.import_key(f)
+
                 ManifestRecursiveLoader(tmp_path / filename,
                                         verify_openpgp=True,
                                         openpgp_env=openpgp_env)
@@ -497,7 +571,7 @@ def test_cli(tmp_path, caplog, manifest_var, key_var, expected):
                               '--no-refresh-keys',
                               '--require-signed-manifest',
                               str(tmp_path)])
-    if str(OpenPGPNoImplementation('')) in caplog.text:
+    if str(OpenPGPNoImplementation('install gpg')) in caplog.text:
         pytest.skip('OpenPGP implementation missing')
 
     eexit = 0 if expected is None else 1
@@ -509,11 +583,16 @@ def test_cli(tmp_path, caplog, manifest_var, key_var, expected):
 EMPTY_DATA = b''
 
 
-@pytest.mark.parametrize('key_var,success', [('VALID_PUBLIC_KEY', True),
-                                             ('MALFORMED_PUBLIC_KEY', False),
-                                             ('EMPTY_DATA', False),
-                                             ('FORGED_PUBLIC_KEY', False),
-                                             ('UNSIGNED_PUBLIC_KEY', False)])
+@pytest.mark.parametrize(
+    'key_var,success',
+    [('VALID_PUBLIC_KEY', True),
+     ('VALID_KEY_NOEMAIL', True),
+     ('VALID_KEY_NONUTF', True),
+     ('MALFORMED_PUBLIC_KEY', False),
+     ('EMPTY_DATA', False),
+     ('FORGED_PUBLIC_KEY', False),
+     ('UNSIGNED_PUBLIC_KEY', False),
+     ])
 def test_env_import_key(openpgp_env, key_var, success):
     """Test importing valid and invalid keys"""
     try:
@@ -528,26 +607,30 @@ def test_env_import_key(openpgp_env, key_var, success):
 
 def test_env_double_close():
     """Test that env can be closed multiple times"""
-    with OpenPGPEnvironment() as env:
+    with IsolatedGPGEnvironment() as env:
         env.close()
 
 
 def test_env_home_after_close():
     """Test that .home can not be referenced after closing"""
-    with OpenPGPEnvironment() as env:
+    with IsolatedGPGEnvironment() as env:
         env.close()
         with pytest.raises(AssertionError):
             env.home
 
 
-@pytest.fixture
-def privkey_env(openpgp_env):
+@pytest.fixture(params=[IsolatedGPGEnvironment,
+                        MockedSystemGPGEnvironment,
+                        ])
+def privkey_env(request):
     """Environment with private key loaded"""
     try:
-        openpgp_env.import_key(io.BytesIO(PRIVATE_KEY))
+        env = request.param()
+        env.import_key(io.BytesIO(PRIVATE_KEY))
     except OpenPGPNoImplementation as e:
         pytest.skip(str(e))
-    return openpgp_env
+    yield env
+    env.close()
 
 
 TEST_STRING = u'The quick brown fox jumps over the lazy dog'
@@ -634,6 +717,23 @@ def test_recursive_manifest_loader_save_submanifest(tmp_path, privkey_env):
     assert m2.openpgp_signature is None
 
 
+@pytest.mark.parametrize(
+    'key_var,expected',
+    [('VALID_PUBLIC_KEY', {KEY_FINGERPRINT: ['gemato@example.com']}),
+     ('OTHER_VALID_PUBLIC_KEY',
+      {OTHER_KEY_FINGERPRINT: ['gemato@example.com']}),
+     ('VALID_KEY_SUBKEY', {KEY_FINGERPRINT: ['gemato@example.com']}),
+     ('VALID_KEY_NOEMAIL', {KEY_FINGERPRINT: []}),
+     ('VALID_KEY_NONUTF', {KEY_FINGERPRINT: ['gemato@example.com']}),
+     ])
+def test_list_keys(openpgp_env_with_refresh, key_var, expected):
+    try:
+        openpgp_env_with_refresh.import_key(io.BytesIO(globals()[key_var]))
+    except OpenPGPNoImplementation as e:
+        pytest.skip(str(e))
+    assert openpgp_env_with_refresh.list_keys() == expected
+
+
 @pytest.fixture(scope='module')
 def global_hkp_server():
     """A fixture that starts a single HKP server instance for tests"""
@@ -650,13 +750,12 @@ def hkp_server(global_hkp_server):
     yield global_hkp_server
 
 
-COMBINED_PUBLIC_KEYS = OTHER_VALID_PUBLIC_KEY + VALID_PUBLIC_KEY
-
-
 REFRESH_VARIANTS = [
     # manifest, key, server key fpr, server key, expected exception
     ('SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', KEY_FINGERPRINT,
      'VALID_PUBLIC_KEY', None),
+    ('SIGNED_MANIFEST', 'VALID_KEY_NONUTF', KEY_FINGERPRINT,
+     'VALID_PUBLIC_KEY', None),
     ('SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', KEY_FINGERPRINT,
      'REVOKED_PUBLIC_KEY', OpenPGPRevokedKeyFailure),
     # test fetching subkey for primary key
@@ -691,7 +790,10 @@ REFRESH_VARIANTS = [
 
 @pytest.mark.parametrize(
     'manifest_var,key_var,server_key_fpr,server_key_var,expected',
-    REFRESH_VARIANTS)
+    REFRESH_VARIANTS +
+    [('SIGNED_MANIFEST', 'VALID_KEY_NOEMAIL', KEY_FINGERPRINT,
+      'VALID_PUBLIC_KEY', None),
+     ])
 def test_refresh_hkp(openpgp_env_with_refresh, hkp_server, manifest_var,
                      key_var, server_key_fpr, server_key_var, expected):
     """Test refreshing against a HKP keyserver"""
@@ -719,12 +821,22 @@ def test_refresh_hkp(openpgp_env_with_refresh, hkp_server, manifest_var,
 
 
 @pytest.mark.parametrize(
-    'manifest_var,key_var,server_key_fpr,server_key_var,expected',
-    REFRESH_VARIANTS)
-def test_refresh_wkd(openpgp_env_with_refresh, manifest_var, key_var,
-                     server_key_fpr, server_key_var, expected):
+    'manifest_var,key_var,server_key_fpr,server_key_var,expected,'
+    'expect_hit',
+    [args + (True,) for args in REFRESH_VARIANTS] +
+    [('SIGNED_MANIFEST', 'VALID_KEY_NOEMAIL', KEY_FINGERPRINT,
+      'VALID_PUBLIC_KEY', OpenPGPKeyRefreshError, False),
+     ])
+def test_refresh_wkd(openpgp_env_with_refresh,
+                     manifest_var,
+                     key_var,
+                     server_key_fpr,
+                     server_key_var,
+                     expected,
+                     expect_hit):
     """Test refreshing against WKD"""
-    with pytest.importorskip('responses').RequestsMock() as responses:
+    with pytest.importorskip('responses').RequestsMock(
+            assert_all_requests_are_fired=expect_hit) as responses:
         try:
             if key_var is not None:
                 with io.BytesIO(globals()[key_var]) as f:
@@ -793,4 +905,47 @@ def test_refresh_wkd_fallback_to_hkp(openpgp_env_with_refresh,
       'iy9q119eutrkn8s1mk4r39qejnbu3n5q?l=Joe.Doe'),
      ])
 def test_get_wkd_url(email, expected):
-    assert OpenPGPEnvironment.get_wkd_url(email) == expected
+    assert get_wkd_url(email) == expected
+
+
+def signal_desc(sig):
+    if hasattr(signal, 'strsignal'):
+        return signal.strsignal(sig)
+    else:
+        return sig
+
+
+@pytest.mark.parametrize(
+    'command,expected,match',
+    [('true', 0, None),
+     ('false', 1, None),
+     ('gpg --verify {tmp_path}/Manifest', 0, None),
+     ('gpg --verify {tmp_path}/Manifest.subkey', 2, None),
+     ('sh -c "kill $$"', -signal.SIGTERM,
+      f'Child process terminated due to signal: '
+      f'{signal_desc(signal.SIGTERM)}'),
+     ('sh -c "kill -USR1 $$"', -signal.SIGUSR1,
+      f'Child process terminated due to signal: '
+      f'{signal_desc(signal.SIGUSR1)}'),
+     ])
+def test_cli_gpg_wrap(tmp_path, caplog, command, expected, match):
+    with open(tmp_path / '.key.bin', 'wb') as f:
+        f.write(VALID_PUBLIC_KEY)
+    with open(tmp_path / 'Manifest', 'w') as f:
+        f.write(SIGNED_MANIFEST)
+    with open(tmp_path / 'Manifest.subkey', 'w') as f:
+        f.write(SUBKEY_SIGNED_MANIFEST)
+
+    command = [x.replace('{tmp_path}', str(tmp_path))
+               for x in shlex.split(command)]
+    retval = gemato.cli.main(['gemato', 'gpg-wrap',
+                              '--openpgp-key',
+                              str(tmp_path / '.key.bin'),
+                              '--no-refresh-keys',
+                              '--'] + command)
+    if str(OpenPGPNoImplementation('install gpg')) in caplog.text:
+        pytest.skip('OpenPGP implementation missing')
+
+    assert retval == expected
+    if match is not None:
+        assert match in caplog.text
